/* GDC -- D front-end for GCC
   Copyright (C) 2004 David Friedman
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
 
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef GCC_DCMPLR_CODEGEN_H
#define GCC_DCMPLR_CODEGEN_H

#include "d-irstate.h"

typedef integer_t xdmd_integer_t;

class ModuleInfo {
public:
    // Assuming only one module is compiled per program instances...
    Array classes; // Array of ClassDeclaration*
    Array ctors; // Arrays of FuncDeclaration*
    Array dtors;
    Array unitTests;
};

typedef enum {
    LIBCALL_ASSERT,
    LIBCALL_ARRAY_BOUNDS,
    LIBCALL_SWITCH_ERROR,
    LIBCALL_INVARIANT,
    LIBCALL_NEWCLASS,
    LIBCALL_NEW,
    LIBCALL_NEWARRAYI,
    LIBCALL_NEWBITARRAY,
    LIBCALL_DELCLASS,
    LIBCALL_DELARRAY,
    LIBCALL_DELMEMORY,
    LIBCALL_CALLFINALIZER,
    LIBCALL_ARRAYSETLENGTH,
    LIBCALL_ARRAYSETLENGTH_B,
    LIBCALL_DYNAMIC_CAST,
    LIBCALL_INTERFACE_CAST,
    LIBCALL_ADEQ,
    LIBCALL_ADEQBIT,
    LIBCALL_ADCMP,
    LIBCALL_ADCMPCHAR,
    LIBCALL_ADCMPBIT,
    LIBCALL_AAIN,
    LIBCALL_AAGET,
    LIBCALL_AAGETRVALUE,
    LIBCALL_AADEL,
    LIBCALL_ARRAYCAST,
    LIBCALL_ARRAYCAST_FROMBIT,
    LIBCALL_ARRAYCOPY,
    LIBCALL_ARRAYCOPYBIT,
    LIBCALL_ARRAYCAT,
    LIBCALL_ARRAYCATB,
    LIBCALL_ARRAYCATN,
    LIBCALL_ARRAYAPPEND,
    LIBCALL_ARRAYAPPENDB,
    LIBCALL_ARRAYAPPENDC,
    LIBCALL_ARRAYAPPENDCB,
    LIBCALL_ARRAYSETBIT,
    LIBCALL_ARRAYSETBIT2,
    LIBCALL_MONITORENTER,
    LIBCALL_MONITOREXIT,
    LIBCALL_CRITICALENTER,
    LIBCALL_CRITICALEXIT,
    LIBCALL_THROW,
    LIBCALL_SWITCH_STRING,
    LIBCALL_SWITCH_USTRING,
    LIBCALL_SWITCH_DSTRING,
    LIBCALL_count
} LibCall;

class ArrayScope;

typedef enum {
    TEnone,
    TEnormal,
    TEall,
    TEprivate,
    TEauto
} TemplateEmission;

// Code generation routines should be in a separate namespace, but so many
// routines need a reference to an IRState to expand Expressions.  Solution
// is to make IRState the code generation namespace with the actual IR state
// routines as a mixin.  There is still the global 'gen' which can be used when
// it is known the full function IR state is not needed. Also a lot of routines
// are static (don't need IR state or expand Expressions), but are still called
// using . or ->.

// Functions without a verb create trees
// Functions with 'do' affect the current instruction stream (or output assembler code.)
// functions with other names are for attribute manipulate, etc.

struct IRState : IRBase
{
    typedef enum  {
	INTRINSIC_BSF, INTRINSIC_BSR,
	INTRINSIC_BT, INTRINSIC_BTC, INTRINSIC_BTR, INTRINSIC_BTS,
	INTRINSIC_BSWAP,
	INTRINSIC_INP, INTRINSIC_INPW, INTRINSIC_INPL,
	INTRINSIC_OUTP, INTRINSIC_OUTPW, INTRINSIC_OUTPL,
	INTRINSIC_STD_VA_ARG,
	INTRINSIC_C_VA_ARG,
	INTRINSIC_C_VA_START,
	INTRINSIC_count,
    } Intrinsic;

    static void doLineNote(const Loc & loc);
    static void setLoc(const Loc & loc);

    // ** Declaration maninpulation
    static void setDeclLoc(tree t, const Loc & loc);

    // Some DMD Declarations don't have the loc set, this searches decl's parents
    // until a valid loc is found.
    static void setDeclLoc(tree t, Dsymbol * decl);

    // Set a DECL's STATIC and EXTERN based on the decl's storage class
    // and if it is to be emitted in this module.
    static void setupSymbolStorage(Dsymbol * decl, tree decl_tree, bool force_static_public = false);
    
    // Definitely in static data, but not neccessarily this module.
    // Assumed to be public data.
    static void setupStaticStorage(Dsymbol * dsym, tree decl_tree);

    // Called just before symbol compilation to handle 'once only' and
    // 'static constructor' attributes.
    static void prepareSymbolOutput(Symbol * s);
    static void outputStaticSymbol(tree t);

    static void addAggMethods(tree rec_type, AggregateDeclaration * agg);

    static void declareType(tree t, Type * d_type);
    static void declareType(tree t, Dsymbol * d_sym);

protected:
    static void declareType(tree t, tree decl);
public:
    
    static tree declContext(Dsymbol * d_sym);
    
    // Hack for systems without linkonce support
    static TemplateEmission emitTemplates;
    static bool shouldEmit(Dsymbol * d_sym);
    static bool shouldEmit(Symbol * sym);

    // ** Local variables
    void emitLocalVar(VarDeclaration * v, bool uninit_okay = false);
    tree localVar(tree t_type);
    tree localVar(Type * e_type) { return localVar(e_type->toCtype()); }
    void expandDecl(tree t_decl);
    
    // ** Type conversion
    
    // 'convertTo' just to give it a different name from the extern "C" convert
    tree convertTo(Expression * exp, Type * target_type);
    tree convertTo(tree exp, Type * exp_type, Type * target_type);
    
    tree convertForAssignment(Expression * exp, Type * target_type);
    tree convertForAssignment(tree exp_tree, Type * exp_type, Type * target_type);

    tree convertForInitialization(Expression * exp, Declaration * target_decl); // handle refs %% needed for constructing init value for _d_newarrayi
    tree convertForInitialization(tree exp_tree, Type * exp_type, Declaration * target_decl);
    tree convertForArgument(Expression * exp, Argument * arg);
protected:
    tree doConvertForInit(Expression * exp, tree in_exp_tree, Type * in_exp_type, Declaration * target_decl);
public:
    tree convertForCondition(Expression * exp) {
	return convertForCondition(exp->toElem(this), exp->type); }
    tree convertForCondition(tree exp_tree, Type * exp_type);
    tree convertForVarArg(Expression * exp);
    tree rawArray(Expression * exp);

    // Can't use VAR_DECLs for the DECL_INITIAL of static varibles or in CONSTRUCTORSs
    static tree stripVarDecl(tree value);

    
    // ** Type management
    static Type * getDType(tree t) {
	// %% TODO: assert that its a type node..
	struct lang_type * l = TYPE_LANG_SPECIFIC( t );
	return l ? l->d_type : 0;
    }

    unsigned getPointerSize()
    {
	return int_size_in_bytes( TREE_TYPE( null_pointer_node ));
    }

    static Type * getObjectType() {
	return ClassDeclaration::classinfo->baseClass->type;
    }
    static Type * getStructInterfaceType() {
	return ((VarDeclaration*) ClassDeclaration::classinfo->fields.data[3])->type/*->toBasetype()*/->next;
    }

    // Routines to handle variables that are references.
    static bool isDeclarationReferenceType(Declaration * decl);
    static tree trueDeclarationType(Declaration * decl);
    static bool isArgumentReferenceType(Argument * arg);
    static tree trueArgumentType(Argument * arg);

    static tree arrayType(Type * d_type, uinteger_t size) // %% use of integer_t
    { return arrayType(d_type->toCtype(), size); }
    static tree arrayType(tree type_node, uinteger_t size);

    // Can't call this until common types have been built
    static bool haveLongDouble() {
	return TYPE_MODE(long_double_type_node) != TYPE_MODE(double_type_node);
    }

    static tree addTypeAttribute(tree type, const char * attrname, tree value = NULL_TREE);
    static void addDeclAttribute(tree type, const char * attrname, tree value = NULL_TREE);

    // ** Simple constants

    static tree nullPointer() { return d_null_pointer; }
    static tree integerConstant(xdmd_integer_t value, tree type = 0);
    static tree integerConstant(xdmd_integer_t value, Type * type) {
	return integerConstant(value, type->toCtype());
    }

    static tree floatConstant( real_t value, TypeBasic * target_type );

    // ** Routines for built in structured types

    static tree realPart(tree c);
    static tree imagPart(tree c);
    
    // ** Dynamic arrays
    static tree darrayPtrRef(tree exp);
           tree darrayPtrRef(Expression * e) { return darrayPtrRef(e->toElem(this)); }
    static tree darrayLenRef(tree exp);
    
    static tree darrayVal(tree type, tree len, tree data);
    // data may be NULL for a null pointer value
    static tree darrayVal(tree type, unsigned len, tree data);
    static tree darrayVal(Type * type, unsigned len, tree data) {
	return darrayVal(type->toCtype(), len, data);
    }
    static tree darrayString(const char * str);
    
    static char * hostToTargetString(char * str, unsigned length, unsigned unit_size);

    // Length of either a static or dynamic array
    tree arrayLength(Expression * exp)
    { return arrayLength(exp->toElem(this), exp->type); }
    static tree arrayLength(tree exp, Type * exp_type);

    // Delegates
    static tree delegateMethodRef(tree exp);
    static tree delegateObjectRef(tree exp);
    static tree delegateVal(tree method_exp, tree object_exp, Type * d_type);
    // These are for references to nested functions/methods as opposed to a variable of
    // type Tdelegate
    tree methodCallExpr(tree callee, tree object, Type * d_type) {
	tree t = delegateVal(callee, object, d_type);
	D_IS_METHOD_CALL_EXPR( t ) = 1;
	return t;
    }
    void extractMethodCallExpr(tree mcr, tree & callee_out, tree & object_out);
    tree objectInstanceMethod(Expression * obj_exp, FuncDeclaration * func, Type * d_type);

    static tree twoFieldType(tree rec_type, tree ft1, tree ft2, Type * d_type = 0);
    static tree twoFieldType(Type * ft1, Type * ft2, Type * d_type = 0);
    static tree twoFieldCtor(tree rec_type, tree f1, tree f2, int storage_class = 0);
    
    // ** Temporaries (currently just SAVE_EXPRs)
    
    // Create a SAVE_EXPR if it is not safe (or effecient*) to reevaluate
    // the expression. * -- Actually can product inefficient code because
    // it prevents cse.
    Expression * maybeMakeTemp(Expression * e);
    tree maybeMakeTemp(tree t);

    // ** Various expressions
    tree toElemLvalue(Expression * e);
    
    static tree aggDeclStaticInit(AggregateDeclaration * agg_decl);

    static tree addressOf(tree exp) {
	d_mark_addressable(exp);
	tree t = build1(ADDR_EXPR, build_pointer_type(TREE_TYPE(exp)), exp);
#if D_NO_TRAMPOLINES
	if (TREE_CODE( exp ) == FUNCTION_DECL)
	    TREE_STATIC( t ) = 1;
#endif
	return t;
    }
    static tree addressOf(Dsymbol *d) {
	return addressOf( d->toSymbol()->Stree );
    }

    /* Like DMD, we append interface data after the ClassInfo struct. This
       requires making the type of classinfo symbol something other than
       (class ClassInfo).  To get a (ClassInfo*), we need to typecast the
       address of the classinfo symbol. */
    static tree classinfoRef(ClassDeclaration * cd) {
	// return cd->toSymbol()->Sspecial.classDecl.classInfoRef;
	tree addr = addressOf(cd);
	TREE_TYPE( addr ) = ClassDeclaration::classinfo->type->toCtype();
	return addr;
    }

    tree pointerIntSum(Expression * ptr_exp, Expression * idx_exp)
    { return pointerIntSum(ptr_exp->toElem(this), idx_exp->toElem(this)); }
    tree pointerIntSum(tree ptr_node, tree idx_exp);

    // DMD allows { void[] a; & a[3]; }
    static tree
    pvoidOkay(tree t) {
	if ( VOID_TYPE_P( TREE_TYPE( TREE_TYPE( t ))) ) {
	    // ::warning("indexing array of void"); 
	    return convert(Type::tuns8->pointerTo()->toCtype(), t);
	} else
	    return t;
    }

    tree boolOp(enum tree_code code, tree a, tree b) {
	return build(code, boolean_type_node, a, b);
    }
    
    tree checkedIndex(Loc loc, tree index, tree upper_bound, bool inclusive);
    tree boundsCond(tree index, tree upper_bound, bool inclusive);
    tree bitAlignCheck(Loc loc, tree index);
    
    // Returns something that can be used as an l-value or an r-value unless
    // the array is of type bit in which case it can only be used as an r-value
    tree arrayElemRef(IndexExp * aer_exp, ArrayScope * aryscp);

    // Returns something that can be used as an l-value.  For bit arrays, also
    // creates expressions needed to modify or select bits

    // Only one of or_mask_out or shift_count_out is needed .. not
    // sure which is better to use...
    tree arrayElemRef(IndexExp * aer_exp, ArrayScope * aryscp,
	tree * and_mask_out, tree * or_mask_out, tree * shift_count_out);
    
    bool isBitArrayAccess(Expression * e) {
	if (e->op == TOKindex) {
	    Type * e1_basetype = e1_basetype = ((IndexExp *) e)->e1->type->toBasetype();
	    // assuming != Taarray is good enough for == Tarray,Tsarray,Tpointer
	    return e1_basetype->ty != Taarray && e1_basetype->next->isbit();
	} else {
	    return false;
	}
    }
    
    static void extractSliceExp(Expression * slice_exp,
	Expression * & array_out, Expression * & lwr_out, Expression * & upr_out) {
	// %% assumes TOKslice // assert(slice_exp->op == TOKslice);
	array_out = ((SliceExp*) slice_exp)->e1;
	lwr_out = ((SliceExp*) slice_exp)->lwr;
	upr_out = ((SliceExp*) slice_exp)->upr;
    }
    typedef enum {
	GS_CreateBounds = 0x1,
	GS_BoundsCheck = 0x2,
	GS_ExtractPointer = 0x4, // return pointer to start of array/pointer
	GS_ExtractDArray = 0x8, // unimpl: return darray with start of array; fails for pointer
	GS_MakePointer = 0xc, // unimpl: return pointer to array[lwr]
	GS_MakeDArray = 010, // unimpl: return array[lwr..upr]
	GS_ProcessMask = 0x1c
    } GrokSliceFlags;
    void grokSlice(Expression * slice_exp, tree & array_out, tree & lwr_out, tree & upr_out, int flags = GS_CreateBounds);

    tree compound(tree a, tree b, tree type = 0) {
	return build(COMPOUND_EXPR, type ? type : TREE_TYPE(b), a, b);
    }
    tree voidCompound(tree a, tree b) { return compound(a, b, void_type_node); }
    tree maybeCompound(tree a, tree b) {
	if (a == NULL_TREE)
	    return b;
	else if (b == NULL_TREE)
	    return a;
	else
	    return build(COMPOUND_EXPR, TREE_TYPE(b), a, b);
    }
    
    // ** Function calls
    tree call(FuncDeclaration * func_decl, Array * args);
    tree call(Expression * expr, Array * arguments);
    tree call(tree callable, tree object, Array * arguments);
    
    tree libCall(LibCall lib_call, unsigned n_args, Expression ** args, tree result_type = 0);
    tree libCall(LibCall lib_call, Array * args, tree result_type = 0);
    tree assertCall(Loc loc, LibCall libcall = LIBCALL_ASSERT);
    static FuncDeclaration * getLibCallDecl(LibCall lib_call);
    static void replaceLibCallDecl(FuncDeclaration * d_decl);
    // This does not perform conversions on the arguments.  This allows
    // arbitrary data to be passed through varargs without going through the
    // usual conversions.
    static tree libCall(LibCall lib_call, unsigned n_args, tree *args, tree force_result_type = 0);

    // GCC 3.3 does not set TREE_SIDE_EFFECTS call by default. GCC 3.4
    // sets it depending on the const/pure attributes of the funcion
    // and the SIDE_EFFECTS flags of the arguments.

    // We don't have a way to indicate const/pure functions in D, so
    // always set this flag
    static tree buildCall(tree type, tree callee, tree args) {
	tree t = build(CALL_EXPR, type, callee, args);
#if D_GCC_VER <= 33
	TREE_SIDE_EFFECTS(t) = 1;
#endif
	return t;
    }

    // Warnings
    static bool warnSignCompare;

    // %%
    static bool originalOmitFramePointer;
    
protected:
    // This can't be static because 
    static tree maybeExpandSpecialCall(tree call_exp);
public:
    tree floatMod(tree a, tree b, Type * d_type);

    tree typeinfoReference(Type * t);
    
    // ** Other
    static bool inGenObjFile;

    static Module * builtinsModule;
    static Module * intrinsicModule;
    static TemplateDeclaration * stdargTemplateDecl;
    static TemplateDeclaration * cstdargStartTemplateDecl;
    static TemplateDeclaration * cstdargArgTemplateDecl;
    static void setBuiltinsModule(Module * mod) { builtinsModule = mod; }
    static void setIntrinsicModule(Module * mod) { intrinsicModule = mod; }
    static void setStdArg(TemplateDeclaration * td) { stdargTemplateDecl = td; }
    static void setCStdArgStart(TemplateDeclaration * td) { cstdargStartTemplateDecl = td; }
    static void setCStdArgArg(TemplateDeclaration * td) { cstdargArgTemplateDecl = td; }
    static bool maybeSetUpBuiltin(Declaration * decl);

    static tree functionPointer(FuncDeclaration * func_decl) { return addressOf(func_decl); }

    void doThunk(tree thunk_decl, tree target_decl, int offset);
    IRState * endFunction();
protected:
    void outputThunk(tree thunk_decl, tree target_decl, int offset);
public:

    static bool functionNeedsChain(FuncDeclaration *f);

    tree doSimpleFunction(Module * mod, const char * name, tree expr, bool static_ctor, bool public_fn = false);
    tree doFunctionToCallFunctions(Module * mod, const char * name, Array * functions);
	
    // Create a tree node to set multiple elements to a single value
    static tree arraySet(tree ptr, tree src, tree count) {
	tree exp = build( (enum tree_code) D_ARRAY_SET_EXPR, void_type_node,
	    ptr, src, count);
	TREE_SIDE_EFFECTS( exp ) = 1;
	TREE_READONLY( exp ) = 1;
	return exp;
    }

    // Create an array of pointers to objects declared by a Declaration
    //   vtbl:        array of Declaration
    //   ci_override: overrides element zero -- handles ClassDeclaration.vtbl.data[0] which
    //                seems to always be Object's ClassDeclaration
    tree vtable(Array * vtbl, ClassDeclaration * ci_override = 0);

    // Returns the D object that was thrown.  Different from the generic exception pointer
    static tree exceptionObject();

    static tree label(Loc loc, Identifier * ident = 0);
    

    // ** Instruction stream manipulation

    // want to inline gcc3 versions...
    void startCond(Statement * stmt, Expression * e_cond);
    void startElse();
    void endCond();
    void startLoop(Statement * stmt);
    void continueHere();
    void exitIfFalse(tree t_cond, bool is_top_cond = 0);
    void exitIfFalse(Expression * e_cond, bool is_top_cond = 0) {
	exitIfFalse(convertForCondition(e_cond), is_top_cond);
    }
    void endLoop();
    void startCase(Statement * stmt, tree t_cond);
    void doCase(tree t_value, tree t_label);
    void endCase(tree t_cond);
    void continueLoop(Identifier * ident);
    void exitLoop(Identifier * ident);
    void startTry(Statement * stmt);
    void startCatches();
    void startCatch(tree t_type);
    void endCatch();
    void endCatches();
    void startFinally();
    void endFinally();
    static void doLabel(tree t_label);
    static void doReturn(tree t_value);
    static void doJump(Statement * stmt, tree t_label);
    void doExp(tree t) { addExp(t); }
    void doExp(Expression * e) { doExp(e->toElem(this)); } // %% should handle volatile...?

    // ** Callback statement evalutation

    static Array stmtExprList;
    
    tree makeStmtExpr(Statement * statement);
    static void retrieveStmtExpr(tree t, Statement ** s_out, IRState ** i_out);
    
    // ** Module info.  Assuming only one module per run of the compiler.
    static ModuleInfo moduleInfo;
    ModuleInfo & getModuleInfo() { return moduleInfo; }

    // ** static constructors (not D static constructors)
    static tree staticCtorList; // not an array since there is at most one

    static void rodc(tree decl, int top_level)
    {
#if D_GCC_VER < 40
	rest_of_decl_compilation(decl, NULL, top_level, 0);
#else
	rest_of_decl_compilation(decl, top_level, 0);
#endif
    }
};

struct BitConfig {
    Type * elemType;
    unsigned shift;
    unsigned mask;
    tree mask_tree;
    tree shift_tree;
    unsigned bitCountToBytes(unsigned bc) {
	return ((bc + mask) >> shift) * elemType->size();
    }
    unsigned bitCountToWords(unsigned bc) {
	return (bc + mask) >> shift;
    }
    void setType(Type * type);
};

extern BitConfig bitConfig;

extern IRState gen;

// Various helpers that need extra state

struct WrappedExp : Expression
{
    tree exp_node;
    WrappedExp(Loc loc, enum TOK op, tree exp_node, Type * type);
    void toCBuffer(OutBuffer *buf);
    elem *toElem(IRState *irs);
};

struct ListMaker {
    tree head;
    tree * ptail;
    ListMaker() : head(NULL_TREE), ptail( & head ) { }
    ListMaker(tree * alt_head) : head(NULL_TREE), ptail( alt_head ) { }
    void chain(tree t) { *ptail = t; ptail = & TREE_CHAIN(t); }
    void cons(tree p, tree v) {
	*ptail = tree_cons(p,v,NULL_TREE);
	ptail = & TREE_CHAIN(*ptail);
    }
    void cons(tree v) { cons(NULL_TREE, v); }
};

class FieldVisitor {
public:
    AggregateDeclaration * aggDecl;
    FieldVisitor(AggregateDeclaration * decl) : aggDecl(decl) { }
    //virtual doField(VarDeclaration * field) = 0;
    virtual void doFields(Array * fields, AggregateDeclaration * agg) = 0;
    virtual void doInterfaces(Array * bases, AggregateDeclaration * agg) = 0;
    void go() { visit(aggDecl); }
    void visit(AggregateDeclaration * decl);
};

class AggLayout : public FieldVisitor {
public:
    tree aggType;
    ListMaker fieldList;
    AggLayout(AggregateDeclaration * ini_agg_decl, tree ini_agg_type) :
	FieldVisitor(ini_agg_decl),
	aggType(ini_agg_type),
	fieldList(& TYPE_FIELDS( aggType )) { }
    void doFields(Array * fields, AggregateDeclaration * agg);
    void doInterfaces(Array * bases, AggregateDeclaration * agg);
    void addField(tree field_decl, unsigned offset);
    void finish();
};

class ArrayMaker {
    Type * elemType;
    int storageClass;
    tree elemList;
    unsigned count;
public:
    ArrayMaker(Type * elem_type, int storage_class);
    ArrayMaker & add(tree value);
    tree finish();
    
};

class ArrayScope {
    VarDeclaration * v;
    IRState * irs;
public:
    ArrayScope(IRState * ini_irs, VarDeclaration * ini_v, const Loc & loc);
    tree setArrayExp(tree e, Type * t);
    tree finish(tree e);
};



#endif


